/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.intelliform;

import static java.util.stream.Collectors.*;

import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.springframework.stereotype.Component;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import de.ozgcloud.common.binaryfile.TempFileUtils;
import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import lombok.RequiredArgsConstructor;

@Component
@RequiredArgsConstructor
class DepositDataMapper {

	// TODO Resolve code duplication (xta-adapter: de.ozgcloud.eingang.xdomea.XMLHelper)
	private static final DocumentBuilder DOCUMENT_BUILDER = createDocumentBuilder();

	private static DocumentBuilder createDocumentBuilder() {
		var documentBuilderFactory = DocumentBuilderFactory.newInstance();
		try {
			documentBuilderFactory.setFeature("http://apache.org/xml/features/disallow-doctype-decl", true);
			return documentBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			throw new TechnicalException("Failed to configure document builder", e);
		}
	}

	public FormData mapToFormData(DepositData depositData) {
		var incomingFileMap = mapDepositAttachmentsToSortedIncomingFiles(depositData);
		var document = parsePrimaryXmlRepresentation(depositData, incomingFileMap);
		var attachmentGroups = findAttachmentGroups(document);

		return mapToFormDataWithRepresentationsAndAttachments(
				getRepresentations(incomingFileMap, getAttachmentFileIds(attachmentGroups)),
				getAttachmentFileGroups(attachmentGroups, incomingFileMap)
		);
	}

	private List<String> getAttachmentFileIds(Map<String, List<String>> attachmentGroups) {
		return attachmentGroups.values().stream()
				.flatMap(Collection::stream)
				.toList();
	}

	private FormData mapToFormDataWithRepresentationsAndAttachments(
			List<IncomingFile> representations,
			List<IncomingFileGroup> attachments
	) {
		return FormData.builder()
				.attachments(attachments)
				.numberOfAttachments(attachments.size())
				.representations(representations)
				.numberOfRepresentations(representations.size())
				.build();
	}

	Map<String, IncomingFile> mapDepositAttachmentsToSortedIncomingFiles(DepositData depositData) {
		var incomingFilesMap = mapDepositAttachmentsToIncomingFiles(depositData);
		var primaryId = depositData.getPrimaryDataAttachmentId();
		incomingFilesMap.putFirst(primaryId, getIncomingFileById(primaryId, incomingFilesMap));
		return incomingFilesMap;
	}

	private LinkedHashMap<String, IncomingFile> mapDepositAttachmentsToIncomingFiles(DepositData depositData) {
		return depositData.getAttachments()
				.stream()
				.collect(Collectors.toMap(
						Attachment::getId,
						this::mapAttachmentToIncomingFile,
						(u, v) -> v,
						LinkedHashMap::new));
	}

	IncomingFile mapAttachmentToIncomingFile(Attachment attachment) {
		var file = TempFileUtils.writeTmpFile(attachment.content);
		return IncomingFile.builder()
				.id(UUID.randomUUID().toString())
				.vendorId(attachment.id)
				.name(attachment.name)
				.contentType(attachment.contentType)
				.size(file.length())
				.file(file)
				.build();
	}

	private Document parsePrimaryXmlRepresentation(DepositData depositData, Map<String, IncomingFile> incomingFileMap) {
		// Expect that the <primaryDataAttachmentId> refers to the XML file
		return parseDocument(
				getIncomingFileById(depositData.getPrimaryDataAttachmentId(), incomingFileMap)
		);
	}

	private static Document parseDocument(IncomingFile incomingFile) {
		try (var inputStream = incomingFile.getContentStream()) {
			return DOCUMENT_BUILDER.parse(inputStream);
		} catch (SAXException | IOException e) {
			throw new TechnicalException("Failed to parse xml document!", e);
		}
	}

	Map<String, List<String>> findAttachmentGroups(Document document) {
		return streamElements(document.getElementsByTagName("file"))
				.collect(groupingBy(
						element -> element.getParentNode().getNodeName(),
						mapping(
								element -> element.getAttribute("id"),
								toList())));
	}

	private List<IncomingFileGroup> getAttachmentFileGroups(Map<String, List<String>> attachmentGroups, Map<String, IncomingFile> incomingFileMap) {
		return attachmentGroups.entrySet().stream()
				.map(entry -> IncomingFileGroup.builder()
						.name(entry.getKey())
						.files(entry.getValue().stream().map(id -> getIncomingFileById(id, incomingFileMap)).toList())
						.build())
				.toList();
	}

	private List<IncomingFile> getRepresentations(Map<String, IncomingFile> incomingFileMap, List<String> attachmentFileIds) {
		return getNamesWithout(incomingFileMap.keySet(), attachmentFileIds).stream()
				.map(id -> getIncomingFileById(id, incomingFileMap))
				.toList();
	}

	private IncomingFile getIncomingFileById(String id, Map<String, IncomingFile> incomingFileMap) {
		if (!incomingFileMap.containsKey(id)) {
			throw new TechnicalException("Failed to find <file> attachment ID '%s' in deposit data!".formatted(id));
		}
		return incomingFileMap.get(id);
	}

	private List<String> getNamesWithout(Collection<String> names, Collection<String> excludedStrings) {
		var excludedStringsSet = Set.copyOf(excludedStrings);
		return names.stream().filter(name -> !excludedStringsSet.contains(name)).toList();
	}

	private Stream<Element> streamElements(NodeList nodeList) {
		return IntStream.range(0, nodeList.getLength())
				.mapToObj(nodeList::item)
				.map(Element.class::cast);
	}

}
