/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.semantik.enginebased.formsolutions;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormDataTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroupTestFactory;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.semantik.enginebased.formsolutions.FormSolutionsFilesMapper.ZippedAttachmentsProcessor;

class FormSolutionsFilesMapperTest {

	private static final String ATTACHMENT_ZIP_FILE_NAME = "attachments.zip";
	private static final String ZIP_CONTENT_TYPE = "application/zip";

	@Spy
	@InjectMocks
	private final FormSolutionsFilesMapper mapper = new FormSolutionsFilesMapper();

	@Nested
	class TestParseFormData {
		@Test
		void shouldCallReadAttachments() {
			mapper.parseFormData(FormDataTestFactory.create());

			verify(mapper).readAttachments(any());
		}

		@Test
		void shouldAddGroup() {
			var attachmentGroup = IncomingFileGroupTestFactory.create();
			doReturn(List.of(attachmentGroup)).when(mapper).readAttachments(any());

			var result = mapper.parseFormData(FormDataTestFactory.create());

			assertThat(result.getAttachments()).containsOnly(attachmentGroup);
		}

		@Test
		void shouldHaveNoAttachmentIfMissing() {
			doReturn(Collections.emptyList()).when(mapper).readAttachments(any());

			var result = mapper.parseFormData(FormDataTestFactory.create());

			assertThat(result.getAttachments()).isEmpty();
		}

	}

	@Nested
	class TestZippedAttachmentsProcessor {

		private final List<IncomingFileGroup> attachments = new ArrayList<>();

		private final ZippedAttachmentsProcessor processor = spy(mapper.new ZippedAttachmentsProcessor(attachments));

		@Nested
		class Process {

			@BeforeEach
			void fillAttachmentList() {
				attachments.add(IncomingFileGroupTestFactory.create());
				when(processor.nonZipFileGroups()).thenReturn(Stream.empty());
			}

			@Test
			void shouldCallExtractAttachments() {
				processor.process();

				verify(processor).extractAttachments();
			}

			@Test
			void shouldReturnExtractedGroups() {
				var expected = IncomingFileTestFactory.create();
				doReturn(Stream.of(expected)).when(processor).extractAttachments();

				var result = processor.process();

				assertThat(result).flatMap(IncomingFileGroup::getFiles).containsOnly(expected);
			}

			@Test
			void shouldReturnEmptyIfNoAttachments() {
				attachments.clear();

				var result = processor.process();

				assertThat(result).isEmpty();
			}
		}

		@Nested
		class ExtractAttachments {

			@Captor
			private ArgumentCaptor<IncomingFile> fileCaptor;

			@Test
			void shouldCallUnzip() {
				attachments.add(buildZipFileGroup());

				processor.extractAttachments().toList();

				verify(processor).unzip(fileCaptor.capture());
			}

			@Test
			void shouldNotCallUnzipForOtherGroup() {
				attachments.add(IncomingFileGroupTestFactory.create());

				processor.extractAttachments().toList();

				verify(processor, never()).unzip(any());
			}

			private IncomingFileGroup buildZipFileGroup() {
				return IncomingFileGroupTestFactory.createBuilder().clearFiles()
						.name(FormSolutionsFilesMapper.FILE_GROUP_ZIP_NAME)
						.file(buildZipFile())
						.build();
			}

		}

		@Nested
		class Unzip {

			@Test
			void shouldCallReadFromZip() {
				IncomingFile zipFile = buildZipFile();

				processor.unzip(zipFile);

				verify(processor).readFromZip(zipFile);
			}

			@Test
			void resultShouldContainUnzippedContent() {
				var resultFile = IncomingFileTestFactory.create();
				doReturn(Stream.of(resultFile)).when(processor).readFromZip(any());

				var result = processor.unzip(buildZipFile());

				assertThat(result).containsOnly(resultFile);
			}

			@Test
			void shouldReturnZipFileOnException() {
				doThrow(new RuntimeException()).when(processor).readFromZip(any());
				IncomingFile zipFile = buildZipFile();

				var result = processor.unzip(zipFile);

				assertThat(result).hasSize(1).contains(zipFile);
			}
		}

		private IncomingFile buildZipFile() {
			return IncomingFileTestFactory.createBuilder()
					.name(ATTACHMENT_ZIP_FILE_NAME).contentType(ZIP_CONTENT_TYPE)
					.build();
		}
	}
}